;
; Disk Access Sub-Routines for FM-R/TOWNS/FM16-121
;   Include from YADISK.ASM
;

;
; Read Physical Block from Disk
;  IN: bx = Physical Block Number (high)
;      dx =     "      "      "    (low)
;      cx = Read Block Count
;      al = SCSI-ID
;      (ah = LUN)
;      [ds:si] = Read Memory Address
; OUT: Carry Set -> Error
;      ax = MS-DOS Status Word
;      cx = Not Red Block Count
RedDsk:
	mov	bh,ah		; Set LUN
	mov	ah,05h		; FM-R Read BIOS Function
	jmp	FMR_RW

;
; Write & Verify Physical Block to Disk
;  IN: bx = Physical Block Number (high)
;      dx =     "      "      "    (low)
;      cx = Read Block Count
;      al = SCSI-ID
;      (ah = LUN)
;      [ds:si] = Read Memory Address
; OUT: Carry Set -> Error
;      ax = MS-DOS Status Word
;      cx = Not Red Block Count
; Destroy -> bh
;
WrtVDsk:
	mov	bh,ah		; Set LUN
	mov	ah,06h		; FM-R Read BIOS Function
	push	bx
	push	ax
	push	cx
	push	dx
	call	FMR_RW		; Write Data to Disk
	jc	WrtVDsk9
	pop	dx
	pop	cx
	pop	ax
	pop	bx
	mov	ah,07h		; FM-R Verify BIOS Function
	jmp	FMR_RW
WrtVDsk9:
	pop	dx
	pop	cx
	pop	bx
	pop	bx
	ret

;
; Write Physical Block from Disk
;  IN: bx = Physical Block Number (high)
;      dx =     "      "      "    (low)
;      cx = Read Block Count
;      al = SCSI-ID
;      (ah = LUN)
;      [ds:si] = Read Memory Address
; OUT: Carry Set -> Error
;      ax = MS-DOS Status Word
;      cx = Not Write Block Count
;
WrtDsk:
	mov	bh,ah		; Set LUN
	mov	ah,06h		; FM-R Write BIOS Function

; Setup FM-R Entry Parameter
;  IN: ah = BIOS Func, al = Device Number
;      ch = 00h, cl = Logical Block Number (Higher)
;      dx = Logical Block Number (Lower)
;      bx = Block Count, ds:di = Buffer Address
; OUT: cx = Not R/W Block Size
;
FMR_RW:
	push	di
	push	si
;
	xchg	cx,bx		; Exchange Block Number & Sector Count
	mov	di,si		; Set Buffer Address
	and	al,00001111b	; Only SCSI-ID
	or	al,10110000b	; Select SCSI Hard Disk
	int	93h		; BIOS Call
	cmp	ah,0		; Error ?
	jz	FMR_RW8

; Check Error Information
	cmp	al,02h		; Illegal Drive Number ?
	jnz	FMR_RW2
	mov	al,1
	jmp	FMR_RW7
FMR_RW2:
	test	cl,05h		; Drive Not Ready & Non-Connect Drive ?
	jz	FMR_RW3
	mov	al,2		; Set Drive Not Ready Error
	jmp	FMR_RW7
FMR_RW3:
	test	cl,0f0h		; Media, Parity, Command Error ?
	jz	FMR_RW4
	mov	al,11		; Set Read Error
	jmp	FMR_RW7
FMR_RW4:
	mov	al,12		; It's Parameter Error
FMR_RW7:
	mov	ah,80h		; Set Error Bit
	stc			; Carry Set
	jmp	FMR_RW9
FMR_RW8:
	xor	ax,ax		; Carry Clear & Status Word Clear
FMR_RW9:
	mov	cx,bx		; Not Read/Write Block Size
	pop	si
	pop	di
	ret

;
; Request Sense
;  IN: al = SCSI-ID
;      ah = LUN
;      [ds:si] = Data Buffer
; OUT: Carry Set-> Error
;
ReqSense:
	push	di
	push	si
	lea	di,[si+16]	; Set Parameter Pointer
; Set Data Buffer Address
	mov	word ptr [di],si
	mov	word ptr [di+2],ds
; Setup CDB for Inquiry Command
	mov	byte ptr [di+4],03h
	mov	byte ptr [di+5],0
	mov	byte ptr [di+6],0
	mov	byte ptr [di+7],0
	mov	byte ptr [di+8],8
	mov	byte ptr [di+9],0
	call	Phase		; Call Secret BIOS Call
	pop	si
	pop	di
	ret

;
; Prevent/Allow Medium
;  IN: al = SCSI-ID
;      ah = LUN
;      cl = Flag Prevent/Allow
; OUT: Carry Set-> Error
;
MedRemov:
	push	di
	push	si
	mov	di,offset TempBuf	; Set Parameter Pointer
; Set Data Buffer Address (None)
	mov	word ptr [di],0000h
	mov	word ptr [di+2],0000h
; Setup CDB for Inquiry Command
	mov	byte ptr [di+4],1eh
	mov	byte ptr [di+5],0
	mov	byte ptr [di+6],0
	mov	byte ptr [di+7],0
	mov	byte ptr [di+8],cl
	mov	byte ptr [di+9],0
	call	Phase		; Call Secret BIOS Call
	pop	si
	pop	di
	ret

;
; Read Capacity (Get Sictor Size)
;  IN: al = SCSI-ID
;      ah = LUN
; OUT: cx = Block Size
;      Carry Set-> Error
;
ReadCap:
	push	dx
	push	bx
	mov	ch,ah		; Set LUN
	mov	ah,02h		; Read Capacity Function
	and	al,00001111b	; Only SCSI-ID
	or	al,10110000b	; Select SCSI Hard Disk
	int	93h		; BIOS Call
	or	ah,ah		; Error ?
	jnz	RedCap8		; Yes, Branch
 	mov	cl,al		; Set Block Size Information
	mov	ax,0080h	; 128/Block
	shl	ax,cl		; Make Block Size ( & Carry Clear)
	mov	cx,ax		; Set Byte/Block Return Value
	pop	bx
	pop	dx
	ret
; Error Return
RedCap8:
	pop	bx
	pop	dx
	jmp	Phase8

;
; FM-R Secret BIOS Call
;  IN: al = SCSI-ID
;      ah = LUN
;      [ds:di+0] = Data Offset
;            +2  = Data Segment
;            +4  = CDB Data 0
;            +5  = CDB Data 1
;             .  .   .  .   n
;
; OUT: ch = Status
;      al = Message (82h = Hardware Error)
;      Carry Set -> if, Status != 0
Phase:
	push	cx
	mov	ch,ah		; Set LUN
	mov	ah,0fdh		; Direct Command Function
	and	al,00001111b	; Only SCSI-ID
	or	al,10110000b	; Select SCSI Hard Disk
	int	93h		; BIOS Call
	mov	ah,ch		; Set Status byte
	pop	cx
	or	ax,ax		; Error ?
	jnz	Phase8		; No, Branch
	clc			; Set Carry on Error
	ret
Phase8:
	mov	ax,8002h	; Drive Not Ready
	stc
	ret

;---------------------- Break Address -----------------------;
Break:

;
; INQUIRY (Check for Removeable Media Drive)
;  IN: al = SCSI-ID
;      ah = LUN
;      [ds:si] = Inquiry Data Buffer
; OUT: Carry Set-> Error
;
Inquiry:
	push	di
	push	si
;
	lea	di,[si+32]
; Set Data Buffer Address
	mov	word ptr [di],si
	mov	word ptr [di+2],ds
; Setup CDB for Inquiry Command
	mov	byte ptr [di+4],012h
	mov	byte ptr [di+5],0
	mov	byte ptr [di+6],0
	mov	byte ptr [di+7],0
	mov	byte ptr [di+8],4
	mov	byte ptr [di+9],0
	call	Phase		; Call Secret BIOS Call
	pop	si
	pop	di
	ret
