	page	60,96
;
; FM-TOWNS Native Mode SCSI-Handler
;         Copyright 1990 by S.Unozawa
;

P_Data equ	0c30h		; SCSI Data Register
P_Cmds equ	0c32h		; SCSI Command Register
P_Stat equ	0c32h		; SCSI Status Register
;
codeseg segment dword 'CODE'
	assume	cs:codeseg
	public	Phase
	db	'Phase',5

;
; Jump Table
P_Table label dword
	dd	offset DataOut
	dd	offset Command
	dd	offset NoDefs
	dd	offset MsgOut
	dd	offset DataIn
	dd	offset Status
	dd	offset NoDefs
	dd	offset MsgIn

;
; SCSI Phase
;  IN: al = SCSI-ID (0 - 6)
;      ds:ebx = CDB Pointer
;      es:edi = Parameter Pointer
; OUT: al = Status
;      ah = Message
;      ecx = Transfer Byte
;
Phase	proc	near
	enter	6,0		; Reserve Message,Status,Count Work
	push	edi
	push	ebx
; Selection Phase
	mov	al,6		;<-- !!!!!! DEBUG FIXED SCSI-ID !!!!!!!!!
	call	Select		; Go Selection Phase
	jc	short Phase4

; Check Current Data Phase
Phase1:
	mov	dx,P_Stat	; Set Status Port Address
	xor	eax,eax		; Clear for Switch Jump
Phase2:
	in	al,dx		; Get Status
	and	al,0f8h		; Only REQ,I/O,MSG,C/D,BSY
	mov	cl,al		; Save Current Status
	shr	al,4		; Check BSY down
	jnc	short Phase8	; Branch BSY down
	test	al,08h		; REQ Set ?
	jz	short Phase2	; Yet Branch
	and	al,07h		; Mask REQ (Only I/O,MSG,C/D)

; Switch Call Each Transfer Phase
	mov	esi,eax		; Set Index
	mov	ah,cl		; Set Current Phase
	mov	edx,offset codeseg:P_Table
	call	[edx+esi*4]	; Go Phase
	jnc	short Phase1	; if Carry Set, it's NoDefs Called

; Hardware Error, return -1
Phase4:
	xor	eax,eax		; Clear Higher 4 bit
	mov	dx,P_Cmds
	out	dx,al		; Reset Command Register
	mov	ecx,eax		; Set Transfer byte 0
	mov	al,0		; Set Status '0'
	mov	ah,0ffh		; Set Message '0ffh'
	pop	ebx
	pop	edi
	leave
	ret

; Reset SCSI Port & Set Return Value
Phase8:
	xor	eax,eax		; Clear Higher 4 bit
	mov	dx,P_Cmds
	out	dx,al		; Reset Command Register
	xor	ecx,ecx
	mov	cx,[ebp-6]	; Get Transfer byte
	mov	al,[ebp-1]	; Get Status
	mov	ah,[ebp-2]	; Get Message
	pop	ebx
	pop	edi
	leave
	ret

;\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
;\             Transfer Phase Entry                     \
;\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
; Data Out/In Phase
DataOut:
	mov	esi,edi		; Set Parameter Address
	call	DatOut		; Go Data Out Phase
	mov	[ebp-6],ecx	; Save Transfer byte
	ret

DataIn:
	push	edi
	call	DatIn		; Go Data In Phase (Non-Blind Mode)
	mov	[ebp-6],ecx	; Save Transfer Byte
	pop	edi
	ret

; Command Phase
Command:
	mov	esi,ebx		; Set CDB Address
	call	DatOut		; Go Data Out Phase
	ret

; Status Phase
Status:
	push	edi
	lea	edi,[ebp-1]	; Set Message byte Address
	call	DatIn		; Go Data In Phase
	pop	edi
	ret

; Message In Phase
MsgIn:
	push	edi
	lea	edi,[ebp-2]	; Set Status byte Address
	call	DatIn		; Go Data In Phase
	pop	edi
	ret

; Not Define this Phase
MsgOut:
NoDefs:
	stc			; Set Carry
	ret

;
Phase	endp

;
; Selection Phase
;  IN: al = SCSI-ID
; OUT: Carry Set -> SCSI-Bus Busy
;
Select proc	near
	push	ebx
	mov	cl,al		; Set Shift Count
	mov	bl,01h		; for SCSI-ID bit
	shl	bl,cl		; Set bit
	mov	ecx,30000	; Wait Loop Count
	cli			; Interrupt Mask
	mov	dx,P_Stat	; SCSI-Status Port
Sel1:
	in	al,dx		; Get Status
	nop			; I/O Recovery wait
	nop
	test	al,08h		; Set BSY bit ?
	jz	short Sel2	; Yes Down, Branch
	loop	Sel1		; Loop Again
	jmp	short Sel8	; Time Out Error

; Asert SCSI-ID
Sel2:
	mov	dx,P_CMDS	; SCSI-Command Port
	mov	al,0		; Initialize SCSI-I/O Port
	out	dx,al
	mov	dx,P_Data	; SCSI-Data Port
	mov	al,bl		; Set SCSI-ID
	out	dx,al		; Set it.
	mov	dx,P_CMDS	; SCSI-Command Port
	mov	al,084h		; Asert Data & SEL
	out	dx,al

; Wait BSY Set
	mov	ecx,30000	; Wait Loop Count
	mov	dx,P_Stat	; SCSI-Status Port
Sel3:
	in	al,dx		; Get Status
	nop			; I/O Recovery wait
	nop
	test	al,08h		; Set BSY bit ?
	jnz	short Sel9	; Yes Set, Branch
	loop	Sel3		; Loop Again

; Selection Fail !!
Sel8:
	sti			; Interrupt Enable
	stc			; Carry Set
	pop	ebx
	ret

; Selection Success
Sel9:
	mov	dx,P_CMDS	; SCSI-Command Port
	mov	al,080h		; Reset SEL
	out	dx,al
	sti			; Interrupt Enable
	clc			; Carry Clear
	pop	ebx
	ret

Select endp

;
; Data Output Phase
;  IN: ds:esi = transfer address
;      ah = Current Phase
; OUT: ecx = Transfer byte
;
DatOut	proc	near
	push	ebx
	cld			; Data Index Increment
	mov	dx,P_Data	; Data Port Address
	mov	bx,P_Stat	; Status Port Address
	mov	bh,dl		; Set Data Port A0-A7 Address
	xor	ecx,ecx		; Counter Clear
	align	4
DatOut0:
	mov	dl,bl		; Set Status Port Address
DatOut1:
	in	al,dx		; Get Current Phase Status
	and	al,0f8h		; Only REQ,I/O,MSG,C/D,BSY
	cmp	al,ah		; Same Phase ?
	jne	short DatOut2	; Branch, another Phase
	mov	dl,bh		; Set DX Data Port Address
; Data Output
	outsb			; Output Data
	inc	ecx		; Count Up
	jmp	short DatOut0	; Skip Next
; Check Current Phase End
DatOut2:
	test	al,08h		; BSY Set ?
	jz	short DatOut9	; Branch, BSY Down
	test	al,80h		; REQ Set ?
	jz	short DatOut1	; Branch, yet
;
DatOut9:
	pop	ebx
	clc			; Carry Clear
	ret
;
DatOut	endp

;
; Data Input Phase
;  IN: es:edi = transfer address
;      ah = Current Phase
; OUT: ecx = Transfer Count
;
DatIn	proc	near
	push	ebx
	cld			; String Index Increment
	mov	dx,P_Data	; Set Data Port Address
	mov	bx,P_Stat	; Set Status Port Address
	mov	bh,dl		; Set A0-A7 Address of Data Port
	xor	ecx,ecx		; Counter Clear
; Tranfer Loop
	align	4
DatIn0:
	mov	dl,bl		; Set DX Status Port
DatIn1:
	in	al,dx		; Get Current Phase Status
	and	al,0f8h		; Only REQ,I/O,MSG,C/D,BSY
	cmp	al,ah		; Same Phase ?
	jne	short DatIn2	; Branch, another Phase
	mov	dl,bh		; Set DX Data Port
; Data Input
	insb			; Input Data & Store
	inc	ecx		; Count up
	jmp	short DatIn0	; Skip Next
; Check Current Phase End
DatIn2:
	test	al,08h		; BSY Set ?
	jz	short DatIn9	; Branch, BSY Down
	test	al,80h		; REQ Set ?
	jz	short DatIn1	; Branch, yet
;
DatIn9:
	pop	ebx
	clc			; Carry Clear
	ret
;
DatIn	endp
;
codeseg ends
	end
