/*
	program	convalg.c

		convert analog palettes from Z's Staff Kid *.alg to *.p0f

		usage :  convalg  [ <*.alg> | <*.p0f> ]

*/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"stella.h"


#define		rdiv(a,b)		(((a)+(b)/2)/(b))


typedef	struct{
			int		h,
					s,
					v;
}		hsvty;


typedef	struct{
			int		r,
					g,
					b;
}		rgbty;


extern	int
		makep0f(
			string	palname
		);
extern	void
		rdpal(
			FILE	*algfile,
			hsvty	pal[]
		),
		torgb(
			hsvty	pal[],
			rgbty	pal2[]
		);
extern	int
		makealg(
			string	palname
		);
extern	void
		tohsv(
			rgbty	pal [],
			hsvty	palx[]
		),
		wrpal(
			FILE	*algfile,
			hsvty	pal[]
		);
extern	string
		chgext(
			string	name,
			string	ext
		);


void
main(
	int		ac,
	char	**ag
){
	int		rc;
	string	palname;

	if(ac==1){
		puts("--- Heart Beat palette utility v1.0 ---");
		puts("");
		puts("usage : convalg [<*.alg> | <*.p0f>] ");
		exit(1);
	}else{
		palname = ag[1];
		if(		strstr(palname,".alg")==NULL
			and	strstr(palname,".p0f")==NULL){
			puts("clearly declair <*.alg> or <*.p0f>");
			exit(1);
		};
	}

	if(strstr(palname,".alg")!=NULL){
		rc = makep0f(palname);
		if(rc){
			puts("Can't make *.p0f ");
		};
	};

	if(strstr(palname,".p0f")!=NULL){
		rc = makealg(palname);
		if(rc){
			puts("Can't make *.alg ");
		};
	};
}


int
makep0f(
	string	palname
){
	int		rc,
			i;
	FILE	*af;
	char	p0fname[80];
	hsvty	palhsv[16];
	rgbty	palrgb[16];
	rgbty	palp0f[16];

	rc = 0;
	af = fopen(palname,"rb");
	if(af!=NULL){
		rdpal (af,palhsv);
		fclose(af);
		torgb(palhsv,palrgb);

		puts("--- analog files are ... ---");
		puts("");
		for(i=0;i<16;++i){
			printf("pal[%2d]  h[%3d] s[%3d] v[%3d]  =  r[%2d] g[%2d] b[%2d] \n",
				i,
				palhsv[i].h,palhsv[i].s,palhsv[i].v,
				palrgb[i].r,palrgb[i].g,palrgb[i].b
			);
		}

		for(i=0;i<16;++i){
			palrgb[i].r *= 0x11;
			palrgb[i].g *= 0x11;
			palrgb[i].b *= 0x11;
		}
		swab((char*)palrgb,(char*)palp0f,16*sizeof(rgbty));

		strcpy(p0fname,palname);
		chgext(p0fname,"p0f");
		af = fopen(p0fname,"wb");
		if(af!=NULL){
			fwrite(palp0f,sizeof(rgbty),16,af);
			fclose(af);
		}else{
			rc = 1;
		}
	}else{
		rc = 1;
	}
	return(rc);
}


void
rdpal(
	FILE	*algfile,
	hsvty	pal[]
){
	int		i;

	pal[0].h = getw(algfile);
	for(i=9;i<=15;++i){
		pal[i].h = getw(algfile);
	}
	pal[0].s = getw(algfile);
	for(i=9;i<=15;++i){
		pal[i].s = getw(algfile);
	}
	pal[0].v = getw(algfile);
	for(i=9;i<=15;++i){
		pal[i].v = getw(algfile);
	}

	pal[8].h = getw(algfile);
	for(i=1;i<=7;++i){
		pal[i].h = getw(algfile);
	}
	pal[8].s = getw(algfile);
	for(i=1;i<=7;++i){
		pal[i].s = getw(algfile);
	}
	pal[8].v = getw(algfile);
	for(i=1;i<=7;++i){
		pal[i].v = getw(algfile);
	}
}


void
torgb(
	hsvty	pal[],
	rgbty	pal2[]
){
	int		imax = 15;
	int		p;
	long	h,s,v,i,f,m,n,k,r,g,b;

	for(p=0;p<16;++p){
		h = pal[p].h;			/* 1/360 */
		s = pal[p].s;			/* 1/100 */
		v = pal[p].v;			/* 1/100 */

		i = h/60;
		f = h-60*i;				/* 1/60 */
		m = v*(100-s);			/* 1/10000 */
		n = v*(6000-s*f);		/* 1/600000 */
		k = v*(6000-s*(60-f));	/* 1/600000 */

		switch((int)i){
			case 0:
				r = rdiv(imax*v,   100L);
				g = rdiv(imax*k,600000L);
				b = rdiv(imax*m, 10000L);
				break;
			case 1:
				r = rdiv(imax*n,600000L);
				g = rdiv(imax*v,   100L);
				b = rdiv(imax*m, 10000L);
				break;
			case 2:
				r = rdiv(imax*m, 10000L);
				g = rdiv(imax*v,   100L);
				b = rdiv(imax*k,600000L);
				break;
			case 3:
				r = rdiv(imax*m, 10000L);
				g = rdiv(imax*n,600000L);
				b = rdiv(imax*v,   100L);
				break;
			case 4:
				r = rdiv(imax*k,600000L);
				g = rdiv(imax*m, 10000L);
				b = rdiv(imax*v,   100L);
				break;
			case 5:
				r = rdiv(imax*v,   100L);
				g = rdiv(imax*m, 10000L);
				b = rdiv(imax*n,600000L);
				break;
		}

		pal2[p].r = (int)r;
		pal2[p].g = (int)g;
		pal2[p].b = (int)b;
	}
}


int
makealg(
	string	palname
){
	int		rc,
			i;
	FILE	*af;
	char	algname[80];
	hsvty	palhsv[16];
	rgbty	palrgb[16];
	rgbty	palp0f[16];

	rc = 0;

	af = fopen(palname,"rb");
	if(af!=NULL){
		fread(palp0f,sizeof(rgbty),16,af);
		fclose(af);

		swab((char*)palp0f,(char*)palrgb,16*sizeof(rgbty));
		for(i=0;i<16;++i){
			palrgb[i].r /= 0x11;
			palrgb[i].g /= 0x11;
			palrgb[i].b /= 0x11;
		}
		tohsv(palrgb,palhsv);

		puts("--- analog files are ... ---");
		puts("");
		for(i=0;i<16;++i){
			printf("pal[%2d]  h[%3d] s[%3d] v[%3d]  =  r[%2d] g[%2d] b[%2d] \n",
				i,
				palhsv[i].h,palhsv[i].s,palhsv[i].v,
				palrgb[i].r,palrgb[i].g,palrgb[i].b
			);
		}

		strcpy(algname,palname);
		chgext(algname,"alg");
		af = fopen(algname,"wb");
		if(af!=NULL){
			wrpal(af,palhsv);
			fclose(af);
		}else{
			rc = 1;
		}
	}else{
		rc = 1;
	}
	return(rc);
}


void
tohsv(
	rgbty	pal [],
	hsvty	palx[]
){
	int		r100,g100,b100,
			h,s,v,
			sr,sg,sb,
			x,
			i;

	for(i=0;i<16;++i){
		r100 = rdiv(100*pal[i].r,15);
		g100 = rdiv(100*pal[i].g,15);
		b100 = rdiv(100*pal[i].b,15);

		v = 0;
		if(v<r100)	v = r100;;
		if(v<g100)	v = g100;;
		if(v<b100)	v = b100;;

		x = 100;
		if(r100<x)	x = r100;;
		if(g100<x)	x = g100;;
		if(b100<x)	x = b100;;

		if(v!=0){
			s = rdiv(100*(v-x),v);
			if(s!=0){
				sr = (v-r100);			/* 1 / (v-x)	*/
				sg = (v-g100);			/* 1 / (v-x)	*/
				sb = (v-b100);			/* 1 / (v-x)	*/

				select(v){
					caseof(v==r100)
						if(x==g100)	h = (int)rdiv(60*(5L*(v-x)+sb),(v-x));
							else	h = (int)rdiv(60*(1L*(v-x)-sg),(v-x));
					caseof(v==g100)
						if(x==b100)	h = (int)rdiv(60*(1L*(v-x)+sr),(v-x));
							else	h = (int)rdiv(60*(3L*(v-x)-sb),(v-x));
					caseof(v==b100)
						if(x==r100)	h = (int)rdiv(60*(3L*(v-x)+sg),(v-x));
							else	h = (int)rdiv(60*(5L*(v-x)-sr),(v-x));
					;
				}
				if(h==360)	h = 0;;
			}else{
				h = 0;
			}
		}else{
			s = 0;
			h = 0;
		}

		palx[i].h = h;
		palx[i].s = s;
		palx[i].v = v;
	}
}


void
wrpal(
	FILE	*algfile,
	hsvty	pal[]
){
	int		i;

	putw(pal[0].h,algfile);
	for(i=9;i<=15;++i){
		putw(pal[i].h,algfile);
	}
	putw(pal[0].s,algfile);
	for(i=9;i<=15;++i){
		putw(pal[i].s,algfile);
	}
	putw(pal[0].v,algfile);
	for(i=9;i<=15;++i){
		putw(pal[i].v,algfile);
	}

	putw(pal[8].h,algfile);
	for(i=1;i<=7;++i){
		putw(pal[i].h,algfile);
	}
	putw(pal[8].s,algfile);
	for(i=1;i<=7;++i){
		putw(pal[i].s,algfile);
	}
	putw(pal[8].v,algfile);
	for(i=1;i<=7;++i){
		putw(pal[i].v,algfile);
	}
}


string
chgext(
	string	name,
	string	ext
){
	char	*period;

	period      = strchr(name,'.');
	*(period+1) = NULL;
	strcat(name,ext);
	return(name);
}


/*
	end of prog
*/
